
#include "stm32f4xx_conf.h"
#include "ssd1289_config.h"
#include "ssd1289_reg.h"
#include "ssd1289.h"
#include "delay.h"

#define LCD_REG	(*((volatile unsigned short *) 0x6c000000)) // A16 (RS) = L
#define LCD_RAM (*((volatile unsigned short *) 0x6c020000))	// A16 (RS) = H

/* defined in gfx.c */
extern uint16_t _gfx_color;
extern uint16_t _txt_bg_color;
extern uint16_t _txt_fg_color;
extern uint16_t _txt_crsr_x;
extern uint16_t _txt_crsr_y;

typedef struct
{
	uint8_t reg;
	uint16_t data;
	uint8_t wait;

} ssd1289_init_t;

static const ssd1289_init_t ssd1289_inits[] =
{
	{ SSD1289_REG_OSCILLATION,  	0x0001, 50 },	// OSCEN=1
	{ SSD1289_REG_POWER_CTRL_1, 	0xa8a4,  0 },	// fixme DCT=b1010=fosc/4 BT=b001=VGH:+6,VGL:-4 DC=b1010=fosc/4 AP=b010=small to medium
	{ SSD1289_REG_POWER_CTRL_2, 	0x0000,  0 },	// fixme VRC=b100:5.5V
	{ SSD1289_REG_POWER_CTRL_3, 	0x080c,  0 },	// fixme VRH=b1000:Vref*2.165
	{ SSD1289_REG_POWER_CTRL_4, 	0x2b00,  0 },	// fixme VCOMG=1 VDV=b1000:VLCD63*1.05
	{ SSD1289_REG_POWER_CTRL_5, 	0x00b7, 50 },	// fixme nOTP=1 VCM=0x2a:VLCD63*0.77
	{ SSD1289_REG_DRIVER_OUT_CTRL, 	0x293f,  0 },	// RL=1 REV=1 CAD=0 BGR=1 SM=0 TB=1 MUX=0x13f=319
	{ SSD1289_REG_LCD_DRIVE_AC, 	0x0600,  0 },   // FLD=0 ENWS=0 D/C=1 EOR=1 WSMD=0 NW=0x00=0
	{ SSD1289_REG_SLEEP_MODE, 		0x0000,  0 }, 	// SLP=0
	{ SSD1289_REG_ENTRY_MODE, 		0x6038, 50 }, 	// VSMode=0 DFM=3:65k TRAMS=0 OEDef=0 WMode=0 Dmode=0 TY=0 ID=3 AM=0 LG=0
	{ SSD1289_REG_COMPARE_1, 		0x0000,  0 }, 	// CPR=0 CPG=0
	{ SSD1289_REG_COMPARE_2, 		0x0000,  0 }, 	// CPB=0
	{ SSD1289_REG_PORCH_HOR, 		0xef1c,  0 },	// XL=239 HBP=28
	{ SSD1289_REG_PORCH_VERT, 		0x0003,  0 },	// VFP=0 VBP=0
	{ SSD1289_REG_DISPLAY_CTRL, 	0x0233,  0 },	// PT=0 VLE=1 SPT=0 GON=1 DTE=1 CM=0 D=3
	{ SSD1289_REG_FRAME_CYCLE, 		0x0000,  0 },	// fixme (tux: 0x0039) NO=0 SDT=0 EQ=0 DIV=0 SDIV=1 SRTN=1 RTN=9:25 clock
	{ SSD1289_REG_GATE_SCAN_START, 	0x0000, 50 },	// SCN=0
	{ SSD1289_REG_FIRST_WIN_START, 	0x0000,  0 },	// SS1=0
	{ SSD1289_REG_FIRST_WIN_END, 	0x013f,  0 },	// SE1=319
	{ SSD1289_REG_SECND_WIN_START, 	0x0000,  0 },	// SS2=0
	{ SSD1289_REG_SECND_WIN_END, 	0x0000, 50 },	// SE2=0
	{ SSD1289_REG_V_SCROLL_CTRL_1, 	0x0000,  0 },	// VL1=0
	{ SSD1289_REG_V_SCROLL_CTRL_1, 	0x0000,  0 },	// VL2=0
	{ SSD1289_REG_H_RAM_ADR_POS, 	0xef00,  0 },	// HEA=0xef=239 HSA=0
	{ SSD1289_REG_V_RAM_ADR_START, 	0x0000,  0 },	// VSA=0
	{ SSD1289_REG_V_RAM_ADR_END, 	0x013f, 50 },	// VEA=0x13f=319
	{ SSD1289_REG_WR_DATA_MASK_1, 	0x0000,  0 },	// WMR=0 WMG=0
	{ SSD1289_REG_WR_DATA_MASK_2, 	0x0000,  0 },	// WMB=0
	{ SSD1289_REG_GAMMA_CTRL_1, 	0x0707,  0 },	// PKP1=7 PKP0=7
	{ SSD1289_REG_GAMME_CTRL_2, 	0x0204,  0 },	// PKP3=2 PKP2=4
	{ SSD1289_REG_GAMMA_CTRL_3, 	0x0204,  0 },	// PKP5=2 PKP4=2
	{ SSD1289_REG_GAMMA_CTRL_4, 	0x0502,  0 },	// PRP1=5 PRP0=2
	{ SSD1289_REG_GAMMA_CTRL_5, 	0x0507,  0 },	// PKN1=5 PKN0=7 (0x0203)
	{ SSD1289_REG_GAMMA_CTRL_6, 	0x0204,  0 },	// PKN3=2 PNN2=4
	{ SSD1289_REG_GAMMA_CTRL_7, 	0x0204,  0 },	// PKN5=2 PKN4=4
	{ SSD1289_REG_GAMMA_CTRL_8, 	0x0502,  0 },	// PRN1=5 PRN0=2
	{ SSD1289_REG_GAMMA_CTRL_9, 	0x0302,  0 },	// VRP1=3 VRP0=2
	{ SSD1289_REG_GAMMA_CTRL_10, 	0x0302, 50 },	// VRN1=3 VRN0=2
};

/* forward declaration */
void ssd1289_reg_set(const uint8_t reg, const uint16_t val);

static void ssd1289_init_gpio()
{
	/*
	 +--------------------------------------------------------------------------------------------------------------+
	 | GPIOS                                                                                                        |
	 +--------------------------------------------------------------------------------------------------------------+
	 | GLCD pins assignment                                                                                         |
	 +-------------------------+---------------------------+---------------------------+----------------------------+
	 | PA1 <-> RESET <-> RESET | PD0  <-> FSMC_D2  <-> D2  | PE7  <-> FSMC_D4  <-> D4  | PG12  <-> FSMC_NE4 <-> CS  |
	 +-------------------------+ PD1  <-> FSMC_D3  <-> D3  | PE8  <-> FSMC_D5  <-> D5  +----------------------------+
	                           | PD4  <-> FSMC_NOE <-> RD  | PE9  <-> FSMC_D6  <-> D6  |
	                           | PD5  <-> FSMC_NWE <-> WR  | PE10 <-> FSMC_D7  <-> D7  |
	                           | PD8  <-> FSMC_D13 <-> D13 | PE11 <-> FSMC_D8  <-> D8  |
	                           | PD9  <-> FSMC_D14 <-> D14 | PE12 <-> FSMC_D9  <-> D9  |
	                           | PD10 <-> FSMC_D15 <-> D15 | PE13 <-> FSMC_D10 <-> D10 |
	                           | PD11 <-> FSMC_A16 <-> RS  | PE14 <-> FSMC_D11 <-> D11 |
	                           | PD14 <-> FSMC_D0  <-> D0  | PE15 <-> FSMC_D12 <-> D12 |
	                           | PD15 <-> FSMC_D1  <-> D1  +---------------------------+
	                           +---------------------------+
	*/

	RCC_AHB1PeriphClockCmd(RCC_AHB1Periph_GPIOA |
						   RCC_AHB1Periph_GPIOD |
						   RCC_AHB1Periph_GPIOE |
						   RCC_AHB1Periph_GPIOG,
						   ENABLE);

	GPIO_PinAFConfig(GPIOD, GPIO_PinSource0, GPIO_AF_FSMC);		// D2
	GPIO_PinAFConfig(GPIOD, GPIO_PinSource1, GPIO_AF_FSMC);		// D3
	GPIO_PinAFConfig(GPIOD, GPIO_PinSource4, GPIO_AF_FSMC);		// NOE -> RD
	GPIO_PinAFConfig(GPIOD, GPIO_PinSource5, GPIO_AF_FSMC);		// NWE -> WR
	GPIO_PinAFConfig(GPIOD, GPIO_PinSource8, GPIO_AF_FSMC);		// D13
	GPIO_PinAFConfig(GPIOD, GPIO_PinSource9, GPIO_AF_FSMC);		// D14
	GPIO_PinAFConfig(GPIOD, GPIO_PinSource10, GPIO_AF_FSMC);	// D15
	GPIO_PinAFConfig(GPIOD, GPIO_PinSource11, GPIO_AF_FSMC);	// A16 -> RS
	GPIO_PinAFConfig(GPIOD, GPIO_PinSource14, GPIO_AF_FSMC);	// D0
	GPIO_PinAFConfig(GPIOD, GPIO_PinSource15, GPIO_AF_FSMC);	// D1

	GPIO_PinAFConfig(GPIOE, GPIO_PinSource7, GPIO_AF_FSMC);		// D4
	GPIO_PinAFConfig(GPIOE, GPIO_PinSource8, GPIO_AF_FSMC);		// D5
	GPIO_PinAFConfig(GPIOE, GPIO_PinSource9, GPIO_AF_FSMC);		// D6
	GPIO_PinAFConfig(GPIOE, GPIO_PinSource10, GPIO_AF_FSMC);	// D7
	GPIO_PinAFConfig(GPIOE, GPIO_PinSource11, GPIO_AF_FSMC);	// D8
	GPIO_PinAFConfig(GPIOE, GPIO_PinSource12, GPIO_AF_FSMC);	// D9
	GPIO_PinAFConfig(GPIOE, GPIO_PinSource13, GPIO_AF_FSMC);	// D10
	GPIO_PinAFConfig(GPIOE, GPIO_PinSource14, GPIO_AF_FSMC);	// D11
	GPIO_PinAFConfig(GPIOE, GPIO_PinSource15, GPIO_AF_FSMC);	// D12

	GPIO_PinAFConfig(GPIOG, GPIO_PinSource12, GPIO_AF_FSMC);	// NE4 -> CS
}

static void ssd1289_gpio_clock_slow()
{
	GPIO_InitTypeDef gpi;
	gpi.GPIO_Speed = GPIO_Speed_2MHz;
	gpi.GPIO_OType = GPIO_OType_PP;
	gpi.GPIO_PuPd  = GPIO_PuPd_NOPULL;
	gpi.GPIO_Mode  = GPIO_Mode_OUT;

	gpi.GPIO_Pin   = GPIO_Pin_1;	// Reset
	GPIO_Init(GPIOA, &gpi);

	gpi.GPIO_Mode  = GPIO_Mode_AF;

	gpi.GPIO_Pin   = GPIO_Pin_0  | GPIO_Pin_1  | GPIO_Pin_4  | GPIO_Pin_5  |
				     GPIO_Pin_8  | GPIO_Pin_9  | GPIO_Pin_10 | GPIO_Pin_11 |
				     GPIO_Pin_14 | GPIO_Pin_15;
	GPIO_Init(GPIOD, &gpi);

	gpi.GPIO_Pin   = GPIO_Pin_7  | GPIO_Pin_8  | GPIO_Pin_9  | GPIO_Pin_10 |
				     GPIO_Pin_11 | GPIO_Pin_12 | GPIO_Pin_13 | GPIO_Pin_14 |
				     GPIO_Pin_15;
	GPIO_Init(GPIOE, &gpi);

	gpi.GPIO_Pin   = GPIO_Pin_12;
	GPIO_Init(GPIOG, &gpi);
}

static void ssd1289_gpio_clock_fast()
{
	GPIO_InitTypeDef gpi;
	gpi.GPIO_Speed = GPIO_Speed_100MHz;
	gpi.GPIO_OType = GPIO_OType_PP;
	gpi.GPIO_PuPd  = GPIO_PuPd_NOPULL;
	gpi.GPIO_Mode  = GPIO_Mode_OUT;

	gpi.GPIO_Pin   = GPIO_Pin_1;	// Reset
	GPIO_Init(GPIOA, &gpi);

	gpi.GPIO_Mode  = GPIO_Mode_AF;

	gpi.GPIO_Pin   = GPIO_Pin_0  | GPIO_Pin_1  | GPIO_Pin_4  | GPIO_Pin_5  |
				     GPIO_Pin_8  | GPIO_Pin_9  | GPIO_Pin_10 | GPIO_Pin_11 |
				     GPIO_Pin_14 | GPIO_Pin_15;
	GPIO_Init(GPIOD, &gpi);

	gpi.GPIO_Pin   = GPIO_Pin_7  | GPIO_Pin_8  | GPIO_Pin_9  | GPIO_Pin_10 |
				     GPIO_Pin_11 | GPIO_Pin_12 | GPIO_Pin_13 | GPIO_Pin_14 |
				     GPIO_Pin_15;
	GPIO_Init(GPIOE, &gpi);

	gpi.GPIO_Pin   = GPIO_Pin_12;
	GPIO_Init(GPIOG, &gpi);
}

static void ssd1289_fsmc_slow()
{
	RCC_AHB3PeriphClockCmd(RCC_AHB3Periph_FSMC, ENABLE);

	FSMC_NORSRAMTimingInitTypeDef nst;
	nst.FSMC_AddressSetupTime      = 0x0f; //max. 0x0f
	nst.FSMC_AddressHoldTime       = 0x0f; //max. 0x0f
	nst.FSMC_DataSetupTime         = 0xff; //max. 0xff
	nst.FSMC_BusTurnAroundDuration = 0x0f; //max. 0x0f
	nst.FSMC_CLKDivision           = 0x0f; //max. 0x0f
	nst.FSMC_DataLatency           = 0x0f; //not used
	nst.FSMC_AccessMode            = FSMC_AccessMode_A;

	FSMC_NORSRAMInitTypeDef nsi;
	nsi.FSMC_Bank                  = FSMC_Bank1_NORSRAM4;
	nsi.FSMC_DataAddressMux        = FSMC_DataAddressMux_Disable;
	nsi.FSMC_MemoryType            = FSMC_MemoryType_SRAM;
	nsi.FSMC_MemoryDataWidth       = FSMC_MemoryDataWidth_16b;
	nsi.FSMC_BurstAccessMode       = FSMC_BurstAccessMode_Disable;
	nsi.FSMC_WaitSignalPolarity    = FSMC_WaitSignalPolarity_Low;
	nsi.FSMC_WrapMode              = FSMC_WrapMode_Disable;
	nsi.FSMC_WaitSignalActive      = FSMC_WaitSignalActive_BeforeWaitState;
	nsi.FSMC_WriteOperation        = FSMC_WriteOperation_Enable;
	nsi.FSMC_WaitSignal            = FSMC_WaitSignal_Disable;
	nsi.FSMC_ExtendedMode          = FSMC_ExtendedMode_Disable;
	nsi.FSMC_WriteBurst            = FSMC_WriteBurst_Disable;
	nsi.FSMC_AsynchronousWait      = FSMC_AsynchronousWait_Disable;
	nsi.FSMC_ReadWriteTimingStruct = &nst;
	nsi.FSMC_WriteTimingStruct     = &nst;
	FSMC_NORSRAMInit(&nsi);

	FSMC_NORSRAMCmd(FSMC_Bank1_NORSRAM4, ENABLE);
}

static void ssd1289_fsmc_fast()
{
	RCC_AHB3PeriphClockCmd(RCC_AHB3Periph_FSMC, ENABLE);

	FSMC_NORSRAMTimingInitTypeDef nst;
	nst.FSMC_AddressSetupTime      = 0x04; //max. 0x0f
	nst.FSMC_AddressHoldTime       = 0x00; //max. 0x0f
	nst.FSMC_DataSetupTime         = 0x04; //max. 0xff
	nst.FSMC_BusTurnAroundDuration = 0x00; //max. 0x0f
	nst.FSMC_CLKDivision           = 0x00; //max. 0x0f
	nst.FSMC_DataLatency           = 0x00; //not used
	nst.FSMC_AccessMode            = FSMC_AccessMode_A;

	FSMC_NORSRAMInitTypeDef nsi;
	nsi.FSMC_Bank                  = FSMC_Bank1_NORSRAM4;
	nsi.FSMC_DataAddressMux        = FSMC_DataAddressMux_Disable;
	nsi.FSMC_MemoryType            = FSMC_MemoryType_SRAM;
	nsi.FSMC_MemoryDataWidth       = FSMC_MemoryDataWidth_16b;
	nsi.FSMC_BurstAccessMode       = FSMC_BurstAccessMode_Disable;
	nsi.FSMC_WaitSignalPolarity    = FSMC_WaitSignalPolarity_Low;
	nsi.FSMC_WrapMode              = FSMC_WrapMode_Disable;
	nsi.FSMC_WaitSignalActive      = FSMC_WaitSignalActive_BeforeWaitState;
	nsi.FSMC_WriteOperation        = FSMC_WriteOperation_Enable;
	nsi.FSMC_WaitSignal            = FSMC_WaitSignal_Disable;
	nsi.FSMC_ExtendedMode          = FSMC_ExtendedMode_Disable;
	nsi.FSMC_WriteBurst            = FSMC_WriteBurst_Disable;
	nsi.FSMC_AsynchronousWait      = FSMC_AsynchronousWait_Disable;
	nsi.FSMC_ReadWriteTimingStruct = &nst;
	nsi.FSMC_WriteTimingStruct     = &nst;
	FSMC_NORSRAMInit(&nsi);

	FSMC_NORSRAMCmd(FSMC_Bank1_NORSRAM4, ENABLE);
}

/* Hardware Reset */
static void ssd1289_reset()
{
	GPIO_SetBits(GPIOA, GPIO_Pin_1);
	delay_ms(20);
	GPIO_ResetBits(GPIOA, GPIO_Pin_1);
	delay_ms(20);
	GPIO_SetBits(GPIOA, GPIO_Pin_1);
	delay_ms(20);
}

void ssd1289_init()
{
	ssd1289_init_gpio();
	ssd1289_gpio_clock_slow();
	ssd1289_fsmc_slow();
	ssd1289_reset();

	/* Configure GLCD register */
	for(uint8_t i = 0; i < (sizeof(ssd1289_inits) / sizeof(ssd1289_inits[0])); i++)
	{
		ssd1289_reg_set(ssd1289_inits[i].reg, ssd1289_inits[i].data);

		if(ssd1289_inits[i].wait)
			delay_ms(ssd1289_inits[i].wait);
	}

	ssd1289_gpio_clock_fast();
	ssd1289_fsmc_fast();

	ssd1289_clear(BLACK);
}

/*********************************************************************************************************
* Function:  void ssd1289_reg_set(uint8_t index, uint16_t val)
* Input       : index - register number
*             : val   - value to be set
* Returns     : none
* Notes		  : Hardware specific.
**********************************************************************************************************/
void ssd1289_reg_set(const uint8_t reg, const uint16_t val)
{
	LCD_REG = reg;
	LCD_RAM = val;
}

/*********************************************************************
* Function    : void ssd1289_cursor_set(uint16_t x, uint16_t y)
* Description : set start address of lcd ram
* Input       : x,y - pixel coordinates
* Return      : none
* Note        : set start address change
********************************************************************/
void ssd1289_cursor_set(const uint16_t x, const uint16_t y)
{
	ssd1289_reg_set(SSD1289_REG_GDDRAM_X_ADDR, y);	// initial settings for the GDDRAM Y address in the address counter (AC).
	ssd1289_reg_set(SSD1289_REG_GDDRAM_Y_ADDR, x);	// initial settings for the GDDRAM X address in the address counter (AC).
	LCD_REG = SSD1289_REG_GDDRAM_DATA;
}

/*********************************************************************
* Function    : void ssd1289_area_set(uint16_t x1, uint16_t y1, uint16_t x2, uint16_t y2)
* Description : set paint area
* Input       : x1,y1,x2,y2 - pixel coordinates
* Return      : none
* Note        : Side Effects: paint area change
********************************************************************/
void ssd1289_area_set(const uint16_t x1, const uint16_t y1, const uint16_t x2, const uint16_t y2)
{
	ssd1289_reg_set(SSD1289_REG_H_RAM_ADR_POS, (y2 << 8) | y1);	// Source RAM address window
	ssd1289_reg_set(SSD1289_REG_V_RAM_ADR_START, x1);			// Gate RAM address window
	ssd1289_reg_set(SSD1289_REG_V_RAM_ADR_END, x2);				// Gate RAM address window
    ssd1289_cursor_set(x1, y1);
}

/*********************************************************************
* Function    : void ssd1289_put_pixel(uint16_t x, uint16_t y)
* Description : puts pixel
* Input       : x,y - pixel coordinates
* Output      : none
* Note        : none
********************************************************************/
void ssd1289_put_pixel(const uint16_t x, const uint16_t y)
{
    ssd1289_cursor_set(x, y);
    LCD_RAM = _gfx_color;
}

void ssd1289_put_pixel_color(const uint16_t x, const uint16_t y, const uint16_t color)
{
    ssd1289_cursor_set(x, y);
    LCD_RAM = color;
}

/*********************************************************************
* Function    : void ssd1289_bar(uint16_t left, uint16_t top, uint16_t right, uint16_t bottom)
* Description : draws rectangle filled with current color
* Input       : left,top - top left corner coordinates,
*             : right,bottom - bottom right corner coordinates
* Output      : none
* Note        :
*  CGRAM map
*  00000...000EF line0
*  00100...001EF line1
*  00200...002EF line2
*	....
*  13F00...13FEF line320
*  yyyxx...yyyxx
*  address x = x
*  address y = y << 8
*  new line y = y + 0x100
********************************************************************/
void ssd1289_bar(const uint16_t left, const uint16_t top, const uint16_t right, const uint16_t bottom, const uint16_t color)
{
	uint16_t x, y;

	ssd1289_area_set(left, top, right, bottom);

    for(y = top; y <= bottom; y++)
	{
        for(x = left; x <= right; x++)
		{
        	LCD_RAM = color;
        }
    }
}

/*********************************************************************
* Function    : void ssd1289_clear(uint16_t color)
* Description : fill display with color
* Input       : color - paint color
* Output      : none
* Note        : none
********************************************************************/
void ssd1289_clear(const uint16_t color)
{
	ssd1289_bar(0, 0, GetMaxX(), GetMaxY(), color);
}

void ssd1289_draw_bitmap_mono(const uint16_t width, const uint16_t height, const uint8_t* data)
{
	ssd1289_area_set(_txt_crsr_x,
			         _txt_crsr_y,
			         _txt_crsr_x + width - 1,
			         _txt_crsr_y + height -1);

	uint8_t pix = 0;
	for(uint16_t y = 0; y < height; y++)
	{
		uint16_t x = 0;
		while(1)
		{
			if(x >= width)
				break;

			if((x % 8) == 0x00)
				pix = *data++;

			if(pix & 0x80)
				LCD_RAM = _txt_fg_color;
			else
				LCD_RAM = _txt_bg_color;

			pix <<= 1;
			x++;
		}
	}
}

void ssd1289_draw_pic(const uint16_t left, const uint16_t top, const uint16_t right, const uint16_t bottom, const uint16_t* data)
{
	ssd1289_area_set(left, top, right, bottom);

	for(uint16_t y = top; y <= bottom; y++)
	{
        for(uint16_t x = left; x <= right; x++)
		{
        	LCD_RAM = *data++;
        }
    }
}
